"""
Global Enum Types for DHCP v4/v6
"""
from enum import IntEnum

#** Variables **#

class Arch(IntEnum):
    """encodes an architecture type per RFC 4578, Section 2.1."""
    INTEL_X86PC       = 0
    NEC_PC98          = 1
    EFI_ITANIUM       = 2
    DEC_ALPHA         = 3
    ARC_X86           = 4
    INTEL_LEAN_CLIENT = 5
    EFI_IA32          = 6
    EFI_BC            = 7
    EFI_XSCALE        = 8
    EFI_X86_64        = 9

class HwType(IntEnum):
    """hardware type as per RFC 2132 and defined by the IANA"""
    Ethernet             = 1
    ExperimentalEthernet = 2
    AmateurRadioAX25     = 3
    ProteonTokenRing     = 4
    Chaos                = 5
    IEEE802              = 6
    ARCNET               = 7
    Hyperchannel         = 8
    Lanstar              = 9
    Autonet              = 10
    LocalTalk            = 11
    LocalNet             = 12
    UltraLink            = 13
    SMDS                 = 14
    FrameRelay           = 15
    ATM                  = 16
    HDLC                 = 17
    FibreChannel         = 18
    ATM2                 = 19
    SerialLine           = 20
    ATM3                 = 21
    MILSTD188220         = 22
    Metricom             = 23
    IEEE1394             = 24
    MAPOS                = 25
    Twinaxial            = 26
    EUI64                = 27
    HIPARP               = 28
    ISO7816              = 29
    ARPSec               = 30
    IPsec                = 31
    Infiniband           = 32
    CAI                  = 33
    WiegandInterface     = 34
    PureIP               = 35

class StatusCode(IntEnum):
    """
    IANA Status Codes for DHCPv6
    https://www.iana.org/assignments/dhcpv6-parameters/dhcpv6-parameters.xhtml#dhcpv6-parameters-5
    """
    # RFC 3315 par. 24..4
    Success       = 0
    UnspecFail    = 1
    NoAddrsAvail  = 2
    NoBinding     = 3
    NotOnLink     = 4
    UseMulticast  = 5
    NoPrefixAvail = 6
    # RFC 5007
    UnknownQueryType = 7
    MalformedQuery   = 8
    NotConfigured    = 9
    NotAllowed       = 10
    # RFC 5460
    QueryTerminated = 11
    # RFC 7653
    DataMissing          = 12
    CatchUpComplete      = 13
    NotSupported         = 14
    TLSConnectionRefused = 15
    # RFC 8156
    AddressInUse               = 16
    ConfigurationConflict      = 17
    MissingBindingInformation  = 18
    OutdatedBindingInformation = 19
    ServerShuttingDown         = 20
    DNSUpdateNotSupported      = 21
    ExcessiveTimeSkew          = 22



"""
DHCPv4 Message Implementation
"""
from ipaddress import IPv4Address
from typing import List, Optional
from typing_extensions import Annotated, Self

from pystructs import *
from pyderive import dataclass

from .enum import MessageType, OpCode, OptionCode
from .options import OptionList, read_option, write_option
from ..enum import HwType

#** Variables **#
__all__ = ['Message']

#: magic cookie to include in DHCP message
MAGIC_COOKIE = bytes([0x63, 0x82, 0x53, 0x63])

#: zeroed ip-address default for message values
ZeroIp = IPv4Address('0.0.0.0')

#** Functions **#

def get_ip(op: OptionCode, options: OptionList) -> Optional[IPv4Address]:
    """translate option value into ipv4-address"""
    option = options.get(op)
    if not option:
        return
    ip = option.value
    if not isinstance(ip, IPv4Address):
        ip = IPv4Address(ip)
    return ip

#** Classes **#

class Header(Struct):
    """DHCPv4 Message Header Group"""
    opcode:       Annotated[OpCode, Wrap[U8, OpCode]]
    hw_type:      Annotated[HwType, Wrap[U8, HwType]]
    hw_length:    U8
    hops:         U8
    message_id:   U32
    seconds:      U16
    flags:        U16
    client_addr:  IPv4
    your_addr:    IPv4
    server_addr:  IPv4
    gateway_addr: IPv4
    hw_addr:      Annotated[bytes, StaticBytes[16]]
    server_name:  Annotated[bytes, StaticBytes[64]]
    boot_file:    Annotated[bytes, StaticBytes[128]]
    magic_cookie: Annotated[bytes, Const[MAGIC_COOKIE]] = MAGIC_COOKIE

@dataclass(slots=True, repr=False)
class Message:
    op:           OpCode
    id:           int
    client_hw:    bytes
    options:      OptionList
    hw_type:      HwType   = HwType.Ethernet
    hops:         int      = 0
    seconds:      int      = 0
    flags:        int      = 0
    client_addr:  Ipv4Type = ZeroIp
    your_addr:    Ipv4Type = ZeroIp
    server_addr:  Ipv4Type = ZeroIp
    gateway_addr: Ipv4Type = ZeroIp 
    server_name:  bytes    = b''
    boot_file:    bytes    = b''
 
    def __repr__(self) -> str:
        cname = self.__class__.__name__
        op    = self.op.name
        htype = self.hw_type.name 
        return f'{cname}(op={op}, id={self.id!r}, hwtype={htype})'
    
    def message_type(self) -> Optional[MessageType]:
        """
        retrieve message-type from options-list (if present)
        """
        option = self.options.get(OptionCode.DHCPMessageType)
        if option is not None:
            return option.value

    def requested_options(self) -> List[OptionCode]:
        """
        retrieve requested parameter options from options-list (if present)
        """
        option = self.options.get(OptionCode.ParameterRequestList)
        return [] if option is None else option.value

    def broadcast_address(self) -> Optional[IPv4Address]:
        """
        retrieve broadcast-address from options-list (if present)
        """
        return get_ip(OptionCode.BroadcastAddress, self.options)

    def requested_address(self) -> Optional[IPv4Address]:
        """
        retrieve requested-address from options-list (if present)
        """
        return get_ip(OptionCode.RequestedIPAddress, self.options)

    def server_identifier(self) -> Optional[IPv4Address]:
        """
        retrieve server-identifier from options-list (if present)
        """
        return get_ip(OptionCode.ServerIdentifier, self.options)

    def encode(self) -> bytes:
        """
        serialize DHCP Message as bytes to send via socket
        """
        ctx   = Context()
        data  = bytearray()
        data += Header(
            opcode=self.op,
            hw_type=self.hw_type,
            hw_length=len(self.client_hw),
            hops=self.hops,
            message_id=self.id,
            seconds=self.seconds,
            flags=self.flags,
            client_addr=self.client_addr,
            your_addr=self.your_addr,
            server_addr=self.server_addr,
            gateway_addr=self.gateway_addr,
            hw_addr=self.client_hw,
            server_name=self.server_name,
            boot_file=self.boot_file,
        ).encode(ctx)
        for option in self.options:
            data += write_option(ctx, option)
        if not any(opt.opcode == OptionCode.End for opt in self.options):
            data += bytes((OptionCode.End, ))
        return bytes(data)

    @classmethod
    def decode(cls, raw: bytes) -> Self:
        """
        deserialize encoded DHCP Message into Message object
        """
        ctx     = Context()
        header  = Header.decode(ctx, raw)
        options = OptionList()
        while raw[ctx.index] not in (0, OptionCode.End):
            option = read_option(ctx, raw)
            options.append(option)
        return cls(
            op=header.opcode,
            id=header.message_id,
            client_hw=header.hw_addr,
            options=options,
            hw_type=header.hw_type,
            hops=header.hops,
            seconds=header.seconds,
            flags=header.flags,
            client_addr=header.client_addr,
            your_addr=header.your_addr,
            gateway_addr=header.gateway_addr,
            server_name=header.server_name,
            boot_file=header.boot_file,
        )



"""
DHCPv6 Enums
"""
from enum import IntEnum

#** Variables **#
__all__ = ['MessageType', 'OptionCode', 'DuidType']

#** Classes **#

class MessageType(IntEnum):
    SOLICIT      = 1
    ADVERTIZE    = 2
    REQUEST      = 3
    CONFIRM      = 4
    RENEW        = 5
    REBIND       = 6
    REPLY        = 7
    RELEASE      = 8
    DECLINE      = 9
    RECONFIGURE  = 10
    INFO_REQUEST = 11
    RELAY_FORW   = 12
    RELAY_REPL   = 13

class OptionCode(IntEnum):
    UNKNOWKN               = 0
    ClientIdentifier       = 1
    ServerIdentifier       = 2
    NonTemporaryAddress    = 3
    TemporaryAddress       = 4
    Address                = 5
    OptionRequest          = 6
    Preference             = 7
    ElapsedTime            = 8
    RelayMessage           = 9
    Authentication         = 11
    ServerUnicast          = 12
    StatusCode             = 13
    RapidCommit            = 14
    UserClass              = 15
    VendorClass            = 16
    VendorInfo             = 17
    InterfaceId            = 18
    ReconfMessage          = 19
    ReconfAccept           = 20
    DNSRecurisveNameServer = 23
    DomainSearchList       = 24
    IdAssocPrefixDeleg     = 25
    IAPrefix               = 26
    InfoRefreshTime        = 32
    SOL_MAX_RT             = 82
    INF_MAX_RT             = 83

class DuidType(IntEnum):
    LinkLayerPlusTime = 1
    EnterpriseNumber  = 2
    LinkLayer         = 3
    UniqueIdentifier  = 4



"""
DHCPv4 Implementation
"""

#** Variables **#
__all__ = [
    'OpCode', 
    'MessageType', 
    'OptionCode',

    'Message',

    'Option',
    'OptionList',
    'OptEnd',
    'OptServerId',
    'OptSubnetMask',
    'OptBroadcast',
    'OptRouter',
    'OptDNS',
    'OptRequestedAddr',
    'OptIPLeaseTime',
    'OptRenwalTime',
    'OptRebindTime',
    'OptMessageType',
    'OptParamRequestList',
    'OptMaxMessageSize',
    'OptClassIdentifier',
    'OptTFTPServerName',
    'OptTFTPServerIP',
    'OptBootFile',
    'OptPXEPathPrefix',
    'OptUserClassInfo',
    'OptClientSystemArch',
    'OptClientNetworkIface',
    'OptClientMachineID',
    'OptEtherBoot',
]

#** Imports **#
from .enum import *
from .message import *
from .options import *



"""
DHCP Unique Identifier (DUID) Implementations (rfc8415 section-11)
"""
from datetime import datetime, timedelta, timezone
from typing import ClassVar
from typing_extensions import Annotated

from pystructs import *

from .enum import DuidType
from ..enum import HwType

#** Variables **#
__all__ = [
    'read_duid',
    'write_duid',

    'DUID',
    'LinkLayerPlusTime',
    'EnterpriseNumber',
    'LinkLayer',
    'UniqueIdentifier',
]

#: codec handler for HwType enum
HwInt = Annotated[HwType, Wrap[U16, HwType]]

#: January 1, 2000 UTC
EPOCH = datetime(year=2000, month=1, day=1, tzinfo=timezone.utc)

#** Functions **#

def read_duid(raw: bytes) -> 'DUID':
    """
    read duid struct from raw bytes
    """
    ctx    = Context()
    duid   = DuidType(decode(ctx, raw, U8))
    dclass = DUIDS.get(duid, None)
    if dclass is None:
        raise ValueError(f'Unsupported DUID: {duid!r}')
    return dclass.decode(ctx, raw)

def write_duid(duid: 'DUID'):
    """
    serialize duid struct into raw bytes
    """
    ctx = Context()
    return encode(ctx, U16, duid.duid) + duid.encode(ctx)

#** Classes **#

class Datetime(datetime):
 
    def __new__(cls, time: int, *args):
        # support standard datetime actions when extra vars are passed
        if args:
            return super().__new__(cls, time, *args)
        # process seconds declaration when parsing data
        dt = (EPOCH + timedelta(seconds=time)).astimezone(tz=None)
        return super().__new__(cls, 
            dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)

    def __int__(self) -> int:
        return int(self.timestamp())

class DUID(Struct):
    duid: ClassVar[DuidType]

class LinkLayerPlusTime(DUID):
    duid = DuidType.LinkLayerPlusTime
    hw_type: HwInt
    time:    Annotated[int, Wrap[U32, Datetime]]
    address: GreedyBytes

class EnterpriseNumber(DUID):
    duid = DuidType.EnterpriseNumber
    en:         U16
    en_contd:   U16
    identifier: GreedyBytes

class LinkLayer(DUID):
    duid = DuidType.LinkLayer
    hw_type: HwInt
    address: GreedyBytes

class UniqueIdentifier(DUID):
    duid = DuidType.UniqueIdentifier
    uuid: StaticBytes[128]

#** Classes **#

#: map of duid enum to valid subclasses
DUIDS = {
    duid.duid:duid
    for duid in globals().values()
    if isinstance(duid, type) and issubclass(duid, DUID) and duid is not DUID
}



"""
DHCPv6 Option Implementations
"""
from typing import ClassVar
from typing_extensions import Annotated, Self

from pystructs import *
from pyderive import dataclass

from .duid import DUID, read_duid, write_duid
from .enum import OptionCode
from ..enum import StatusCode
from ..base import DHCPOption, DHCPOptionList, Microseconds, Seconds

#** Variables **#
__all__ = [
    'read_option',

    'Option',
    'OptionList',
    'OptClientIdentifier',
    'OptServerIdentifier',
    'OptNonTemporaryAddr',
    'OptTemporaryAddress',
    'OptAddress',
    'OptRequestList',
    'OptPreference',
    'OptElapsed',
    'OptRelay',
    'OptAuth',
    'OptUnicast',
    'OptStatusCode',
    'OptIdAssocPrefixDeleg',
    'OptIAPrefix',
]

#: integer codec ot parse option-code
OptionCodeInt = Annotated[OptionCode, Wrap[U16, OptionCode]]

#** Functions **#

def read_option(ctx: Context, raw: bytes) -> 'Option':
    """
    read and deserialize best option-class to match option content
    
    :param ctx: serialization context
    :param raw: raw bytes to deserialized into an option
    :return:    option object
    """
    # parse option-type/value & retrieve best matching subclass
    opt     = OptStruct.decode(ctx, raw)
    oclass  = OPTIONS.get(opt.code, Option)
    # parse option w/ it's own sub-context
    subctx = Context()
    option = oclass.decode(subctx, opt.value)
    if oclass is Option:
        oclass.opcode = opt.code
    return option

def write_option(ctx: Context, option: 'Option') -> bytes:
    """
    write and serialize option-class into raw-bytes
    
    :param ctx: serialization context
    :param raw: option object to serialize
    :return:    serialized bytes
    """
    # serialize option into bytes
    subctx  = Context()
    content = option.encode(subctx)
    return OptStruct(option.opcode, content).encode(ctx)

#** Classes **#

class OptStruct(Struct):
    code:  OptionCodeInt
    value: Annotated[bytes, SizedBytes[U16]]

class Option(DHCPOption):
    """DHCPv6 BaseClass Option Definition"""
    opcode: ClassVar[OptionCode] = OptionCode.UNKNOWKN

class OptionList(DHCPOptionList):
    """DHCPv6 DHCPOptionList Implementation"""
    pass

class _DuidOption(Option):
    """BaseClass to Support DUID Option Implementations"""
    opcode: ClassVar[OptionCode]
    value:  DUID

    def encode(self, ctx: Context) -> bytes:
        return write_duid(self.value)

    @classmethod
    def decode(cls, ctx: Context, value: bytes) -> Self:
        return cls(read_duid(value))

class OptClientIdentifier(_DuidOption):
    opcode = OptionCode.ClientIdentifier

class OptServerIdentifier(_DuidOption):
    opcode = OptionCode.ServerIdentifier

class OptNonTemporaryAddr(Option):
    opcode = OptionCode.NonTemporaryAddress
    value:   U32
    t1:      U32
    t2:      U32
    options: GreedyBytes

class OptTemporaryAddress(Option):
    opcode = OptionCode.TemporaryAddress
    value:   U32
    options: GreedyBytes

class OptAddress(Option):
    opcode = OptionCode.Address
    value:          IPv6
    pref_lifetime:  Annotated[Seconds, Wrap[U32, Seconds]]
    valid_lifetime: Annotated[Seconds, Wrap[U32, Seconds]]
    options:        GreedyBytes

class OptRequestList(Option):
    opcode = OptionCode.OptionRequest
    value: GreedyList[OptionCodeInt]

class OptPreference(Option):
    opcode = OptionCode.Preference
    value: U8

class OptElapsed(Option):
    opcode = OptionCode.ElapsedTime
    value: Annotated[Microseconds, Wrap[U16, Microseconds]]

class OptRelay(Option):
    opcode = OptionCode.RelayMessage
    value: GreedyBytes

class OptAuth(Option):
    value: ClassVar[int] # remove value as init-arg
    opcode = OptionCode.Authentication
    protocol:   U8
    algorithm:  U8
    rdm:        U8
    replay_det: StaticBytes[8]
    info:       GreedyBytes

class OptUnicast(Option):
    opcode = OptionCode.ServerUnicast
    value: IPv6

class OptStatusCode(Option):
    opcode = OptionCode.StatusCode
    value:   Annotated[StatusCode, Wrap[U16, StatusCode]]
    message: GreedyBytes

class OptIdAssocPrefixDeleg(Option):
    opcode = OptionCode.IdAssocPrefixDeleg
    value:   U32
    t1:      Annotated[Seconds, Wrap[U32, Seconds]]
    t2:      Annotated[Seconds, Wrap[U32, Seconds]]
    options: GreedyBytes
 
    def read_options(self) -> OptionList:
        ctx = Context()
        options = OptionList()
        while ctx.index < len(self.options):
            option = read_option(ctx, self.options)
            options.append(option)
        return options

class OptIAPrefix(Option):
    value: ClassVar[int]
    opcode = OptionCode.IAPrefix
    pref_lifetime:  Annotated[Seconds, Wrap[U32, Seconds]]
    valid_lifetime: Annotated[Seconds, Wrap[U32, Seconds]]
    prefix_length:  U8
    ipv6_prefix:    IPv6
    options:        GreedyBytes

#** Init **#

#: map of option-codes to option-subclasses
OPTIONS = {
    value.opcode:value
    for name, value in globals().items()
    if not name.startswith('_') and isinstance(value, type) and issubclass(value, Option) 
}



"""
DHCPv4 Enums
"""
from enum import IntEnum

#** Variables **#
__all__ = ['OpCode', 'MessageType', 'OptionCode']

#** Classes **#

class OpCode(IntEnum):
    """Message Operation-Code (REQUEST/REPLY)"""
    BootRequest = 1
    BootReply   = 2

class MessageType(IntEnum):
    """DHCP Message Types - DISCOVER, OFFER, etc"""
    Discover  = 1
    Offer     = 2
    Request   = 3
    Decline   = 4
    Ack       = 5
    Nak       = 6
    Release   = 7
    Inform    = 8

class OptionCode(IntEnum):
    """DHCP Request Paramter OptionCodes"""
    OptionPad                                  = 0
    SubnetMask                                 = 1
    TimeOffset                                 = 2
    Router                                     = 3
    TimeServer                                 = 4
    NameServer                                 = 5
    DomainNameServer                           = 6
    LogServer                                  = 7
    QuoteServer                                = 8
    LPRServer                                  = 9
    ImpressServer                              = 10
    ResourceLocationServer                     = 11
    HostName                                   = 12
    BootFileSize                               = 13
    MeritDumpFile                              = 14
    DomainName                                 = 15
    SwapServer                                 = 16
    RootPath                                   = 17
    ExtensionsPath                             = 18
    IPForwarding                               = 19
    NonLocalSourceRouting                      = 20
    PolicyFilter                               = 21
    MaximumDatagramAssemblySize                = 22
    DefaultIPTTL                               = 23
    PathMTUAgingTimeout                        = 24
    PathMTUPlateauTable                        = 25
    InterfaceMTU                               = 26
    AllSubnetsAreLocal                         = 27
    BroadcastAddress                           = 28
    PerformMaskDiscovery                       = 29
    MaskSupplier                               = 30
    PerformRouterDiscovery                     = 31
    RouterSolicitationAddress                  = 32
    StaticRoutingTable                         = 33
    TrailerEncapsulation                       = 34
    ArpCacheTimeout                            = 35
    EthernetEncapsulation                      = 36
    DefaulTCPTTL                               = 37
    TCPKeepaliveInterval                       = 38
    TCPKeepaliveGarbage                        = 39
    NetworkInformationServiceDomain            = 40
    NetworkInformationServers                  = 41
    NTPServers                                 = 42
    VendorSpecificInformation                  = 43
    NetBIOSOverTCPIPNameServer                 = 44
    NetBIOSOverTCPIPDatagramDistributionServer = 45
    NetBIOSOverTCPIPNodeType                   = 46
    NetBIOSOverTCPIPScope                      = 47
    XWindowSystemFontServer                    = 48
    XWindowSystemDisplayManger                 = 49
    RequestedIPAddress                         = 50
    IPAddressLeaseTime                         = 51
    OptionOverload                             = 52
    DHCPMessageType                            = 53
    ServerIdentifier                           = 54
    ParameterRequestList                       = 55
    Message                                    = 56
    MaximumDHCPMessageSize                     = 57
    RenewTimeValue                             = 58
    RebindingTimeValue                         = 59
    ClassIdentifier                            = 60
    ClientIdentifier                           = 61
    NetWareIPDomainName                        = 62
    NetWareIPInformation                       = 63
    NetworkInformationServicePlusDomain        = 64
    NetworkInformationServicePlusServers       = 65
    TFTPServerName                             = 66
    BootfileName                               = 67
    MobileIPHomeAgent                          = 68
    SimpleMailTransportProtocolServer          = 69
    PostOfficeProtocolServer                   = 70
    NetworkNewsTransportProtocolServer         = 71
    DefaultWorldWideWebServer                  = 72
    DefaultFingerServer                        = 73
    DefaultInternetRelayChatServer             = 74
    StreetTalkServer                           = 75
    StreetTalkDirectoryAssistanceServer        = 76
    UserClassInformation                       = 77
    SLPDirectoryAgent                          = 78
    SLPServiceScope                            = 79
    RapidCommit                                = 80
    FQDN                                       = 81
    RelayAgentInformation                      = 82
    InternetStorageNameService                 = 83
    # Option 84 returned in RFC 3679
    NDSServers                       = 85
    NDSTreeName                      = 86
    NDSContext                       = 87
    BCMCSControllerDomainNameList    = 88
    BCMCSControllerIPv4AddressList   = 89
    Authentication                   = 90
    ClientLastTransactionTime        = 91
    AssociatedIP                     = 92
    ClientSystemArchitectureType     = 93
    ClientNetworkInterfaceIdentifier = 94
    LDAP                             = 95
    # Option 96 returned in RFC 3679
    ClientMachineIdentifier     = 97
    OpenGroupUserAuthentication = 98
    GeoConfCivic                = 99
    IEEE10031TZString           = 100
    ReferenceToTZDatabase       = 101
    # Options 102-111 returned in RFC 3679
    Ipv6OnlyPreferred          = 108 # [RFC8925]
    OPTION_DHCP4O6_S46_SADDR   = 109 # [RFC8539]
    NetInfoParentServerAddress = 112
    NetInfoParentServerTag     = 113
    URL                        = 114
    # Option 115 returned in RFC 3679
    AutoConfigure                   = 116
    NameServiceSearch               = 117
    SubnetSelection                 = 118
    DNSDomainSearchList             = 119
    SIPServers                      = 120
    ClasslessStaticRoute            = 121
    CCC                             = 122
    GeoConf                         = 123
    VendorIdentifyingVendorClass    = 124
    VendorIdentifyingVendorSpecific = 125
    # Options 126-127 returned in RFC 3679
    TFTPServerIPAddress                   = 128
    CallServerIPAddress                   = 129
    DiscriminationString                  = 130
    RemoteStatisticsServerIPAddress       = 131
    O_8021PVLANID                         = 132
    O_8021QL2Priority                     = 133
    DiffservCodePoint                     = 134
    HTTPProxyForPhoneSpecificApplications = 135
    PANAAuthenticationAgent               = 136
    LoSTServer                            = 137
    CAPWAPAccessControllerAddresses       = 138
    OPTIONIPv4AddressMoS                  = 139
    OPTIONIPv4FQDNMoS                     = 140
    SIPUAConfigurationServiceDomains      = 141
    OPTIONIPv4AddressANDSF                = 142
    OPTIONIPv6AddressANDSF                = 143
    # Options 144-149 returned in RFC 3679
    GeoLoc                   = 144 #[RFC6225]
    FORCERENEW_NONCE_CAPABLE = 145 #[RFC6704]
    RDNSSSelection           = 146 #[RFC6731]
    OPTION_V4_DOTS_RI        = 147 #[RFC8973]
    OPTION_V4_DOTS_ADDRESS   = 148 #[RFC8973]
    TFTPServerAddress        = 150
    StatusCode               = 151
    BaseTime                 = 152
    StartTimeOfState         = 153
    QueryStartTime           = 154
    QueryEndTime             = 155
    DHCPState                = 156
    DataSource               = 157
    # Options 158-174 returned in RFC 3679
    OPTION_V4_PCP_SERVER             = 158 #[RFC7291]
    OPTION_V4_PORTPARAMS             = 159 #[RFC7618]
    OPTION_MUD_URL_V4                = 161 #[RFC8520]
    OPTION_V4_DNR                    = 162 #[RFC-ietf-add-dnr-13] 
    Etherboot                        = 175
    IPTelephone                      = 176
    EtherbootPacketCableAndCableHome = 177
    # Options 178-207 returned in RFC 3679
    HPDMServer           = 202 #[HP Device-Manager 4.7]
    HPDMGateway          = 203 #[HP Device-Manager 4.7]
    PXELinuxMagicString  = 208
    PXELinuxConfigFile   = 209
    PXELinuxPathPrefix   = 210
    PXELinuxRebootTime   = 211
    OPTION6RD            = 212
    OPTIONv4AccessDomain = 213
    # Options 214-219 returned in RFC 3679
    SubnetAllocation        = 220
    VirtualSubnetAllocation = 221
    # Options 222-223 returned in RFC 3679
    # Options 224-254 are reserved for private use
    FortinetManagerIP      = 240 #[FortiNet FortiManager]
    FortinetManagerDomain  = 241 #[FortiNet FortiManager]
    MSClasslessStaticRoute = 249
    MSEncodingLongOption   = 250
    ProxyAutoDiscover      = 252 #[RFC-draft-ietf-wrec-wpad-01]
    End                    = 255



"""
DHCPv4 Implementation
"""

#** Variables **#
__all__ = [
    'DUID',
    'LinkLayerPlusTime',
    'EnterpriseNumber',
    'LinkLayer',
    'UniqueIdentifier',

    'MessageType', 
    'OptionCode', 
    'DuidType',

    'Message',
    
    'Option',
    'OptionList',
    'OptClientIdentifier',
    'OptServerIdentifier',
    'OptNonTemporaryAddr',
    'OptTemporaryAddress',
    'OptAddress',
    'OptRequestList',
    'OptPreference',
    'OptElapsed',
    'OptRelay',
    'OptAuth',
    'OptUnicast',
    'OptStatusCode',
    'OptIdAssocPrefixDeleg',
    'OptIAPrefix',
]

#** Imports **#
from .duid import *
from .enum import *
from .message import *
from .options import *



"""
Universal DHCP Exceptions based on IANA StatusCodes
"""
from typing import Dict, Type, Any, Optional

from .enum import StatusCode

#** Variables **#
__all__ = [
    'make_error',

    'DhcpError',
    'NoAddrsAvailable',
    'NotAllowed',
    'NotSupported',
    'MalformedQuery',
    'Terminated',
    'UnknownQueryType',
]

_EXCEPTIONS: Dict[StatusCode, Type[Exception]] = {}

#** Functions **#

def make_error(code: StatusCode, message: Any = None):
    """
    retrieve best exception object to match the given rcode
    """
    # cache map of exceptions in module
    global _EXCEPTIONS
    if not _EXCEPTIONS:
        for value in globals().values():
            if isinstance(value, type) and issubclass(value, DhcpError):
                _EXCEPTIONS[value.code] = value
    # retrieve best-matching exception class based on rcode
    eclass = _EXCEPTIONS.get(code, DhcpError)
    raise eclass(message, code)

#** Classes **#

class DhcpError(Exception):
    code: StatusCode = StatusCode.UnspecFail

    def __init__(self, msg: Any = None, code: Optional[StatusCode] = None):
        self.message = msg
        self.code    = code or self.code

    def __str__(self) -> str:
        if self.message and self.__class__.code == self.code:
            return str(self.message)
        return super().__str__()

class NoAddrsAvailable(DhcpError):
    code = StatusCode.NoAddrsAvail

class NotAllowed(DhcpError):
    code = StatusCode.NotAllowed

class NotSupported(DhcpError):
    code = StatusCode.NotSupported

class MalformedQuery(DhcpError):
    code = StatusCode.MalformedQuery

class UnknownQueryType(DhcpError):
    code = StatusCode.UnknownQueryType

class Terminated(DhcpError):
    code = StatusCode.QueryTerminated



"""
DHCPv4 Server Implementation
"""

#** Variables **#
__all__ = [
    'pxe_handler',
    'ipv4_handler',

    'AssignFunc',
    'Assignment',
    'PxeTftpConfig',
    'PxeDynConfig',
    'PxeConfig',
    'Cache',

    'HandlerFunc',
    'Context',
    'Session',
    'SimpleSession'
]

#** Imports **#
from .server import *
from .handlers import *



"""
DHCPv6 Message Implementation
"""
from ipaddress import IPv6Address
from typing import Union
from typing_extensions import Annotated, Self

from pystructs import *
from pyderive import dataclass, field, asdict

from .enum import MessageType
from .options import OptionList, read_option

#** Variables **#
__all__ = ['Message']

#: codec type to support MessageType
MessageInt = Annotated[MessageType, Wrap[U8, MessageType]]

#** Functions **#

def read_options(ctx: Context, raw: bytes) -> OptionList:
    """read options from raw-bytes and return options-list"""
    options = OptionList()
    while ctx.index < len(raw):
        option = read_option(ctx, raw)
        options.append(option)
    return options

def write_options(ctx: Context, data: bytearray, options: OptionList):
    """serialize options into bytearray"""
    for option in options:
        data += option.encode(ctx)

#** Classes **#

class MsgHeader(Struct):
    op: MessageInt
    id: U24

class RelayForwardHeader(Struct):
    op:        MessageInt
    hops:      U16
    link_addr: IPv6
    peer_addr: IPv6

class RelayReplyHeader(Struct):
    op:        MessageInt
    link_addr: IPv6
    peer_addr: IPv6

@dataclass(slots=True, repr=False)
class Message:
    op:      MessageType
    id:      int
    options: OptionList = field(default_factory=OptionList)
 
    def __repr__(self) -> str:
        cname = self.__class__.__name__
        return f'{cname}(op={self.op.name}, id=0x{self.id:02x})'

    def encode(self) -> bytes:
        """
        Encode DHCPv6 client/server message into bytes
        """
        ctx   = Context()
        data  = bytearray()
        data += MsgHeader(self.op, self.id).encode(ctx)
        write_options(ctx, data, self.options)
        return bytes(data)

    @classmethod
    def decode(cls, raw: bytes) -> Self:
        """
        Decode DHCPv6 client/server message from bytes
        """
        ctx     = Context()
        header  = MsgHeader.decode(ctx, raw)
        options = read_options(ctx, raw)
        return cls(header.op, header.id, options)

@dataclass(slots=True, repr=False)
class RelayReplyMessage:
    op:        MessageType
    link_addr: Ipv6Type
    peer_addr: Ipv6Type
    options:   OptionList 

    def __repr__(self) -> str:
        cname  = self.__class__.__name__
        values = ' '.join(f'{k}={v}' for k, v in {
            'op':  self.op.name,
            'hops':getattr(self, 'hops', None), 
            'link': self.link_addr, 
            'peer': self.peer_addr
        }.items() if v is not None)
        return f'{cname}({values})'

    def encode(self) -> bytes:
        """
        Encode DHCPv6 RelayReply message into bytes
        """
        fields = asdict(self)
        fields.pop('options', None)
        # serialize content
        ctx   = Context()
        data  = bytearray()
        data += RelayReplyHeader(**fields).encode(ctx)
        write_options(ctx, data, self.options)
        return bytes(data)
 
    @classmethod
    def decode(cls, raw: bytes) -> Self:
        """
        Decode DHCPv6 RelayReply message from bytes
        """
        ctx     = Context()
        header  = RelayReplyHeader.decode(ctx, raw)
        fields  = asdict(header) #type: ignore
        options = read_options(ctx, raw)
        return cls(options=options, **fields)

@dataclass(slots=True, repr=False)
class RelayForwardMessage(RelayReplyMessage):
    op:        MessageType
    hops:      int
    link_addr: Ipv6Type
    peer_addr: Ipv6Type
    options:   OptionList

    def encode(self) -> bytes:
        """
        Encode DHCPv6 RelayForward message into bytes
        """
        ctx   = Context()
        data  = bytearray()
        data += RelayForwardHeader(self.op, 
            self.hops, self.link_addr, self.peer_addr).encode(ctx)
        write_options(ctx, data, self.options)
        return bytes(data)

    @classmethod
    def decode(cls, raw: bytes) -> Self:
        """
        Decode DHCPv6 RelayForward message from bytes
        """
        ctx     = Context()
        header  = RelayForwardHeader.decode(ctx, raw)
        options = read_options(ctx, raw)
        return cls(
            op=header.op, 
            hops=header.hops, 
            link_addr=header.link_addr, 
            peer_addr=header.peer_addr, 
            options=options
        )



"""
PyServe Session DHCPv4 Implementation
"""
from abc import ABC, abstractmethod
from ipaddress import IPv4Address
from logging import Logger, getLogger
from typing import Callable, List, NamedTuple, Optional

from pyserve import Address, UdpWriter
from pyserve import Session as BaseSession
from pyderive import dataclass, field

from ..enum import MessageType, OpCode, OptionCode
from ..message import IpType, Message
from ..options import *
from ...enum import StatusCode
from ...exceptions import DhcpError, NotAllowed, UnknownQueryType

#** Variables **#
__all__ = ['HandlerFunc', 'Context', 'Session', 'SimpleSession']

#: dhcp response port
PORT = 68

#: broadcast request for responding to messages
BROADCAST = IPv4Address('255.255.255.255')

#: required opcodes to preserve even when not requested
REQUIRED = {
    OptionCode.DHCPMessageType, 
    OptionCode.ServerIdentifier,
    OptionCode.IPAddressLeaseTime,
}

#: function definition for session processing callback
HandlerFunc = Callable[['Context'], None]

#: logger generation function
logger = getLogger('pydhcp.v4')

#** Functions **#

def is_zero(host: IpType) -> bool:
    """check if host-ip is zeroed"""
    return len(IPv4Address(host).packed.strip(b'\x00')) == 0

def find_host(*ips: IpType) -> IPv4Address:
    """search through available ips to find first non-zero available"""
    for ip in (IPv4Address(ip) for ip in ips):
        if not is_zero(ip):
            return ip
    raise RuntimeError(f'No NonZero IPs: {ips!r}')

def mac_address(hwaddr: bytes) -> str:
    """translate hardware address to mac"""
    return ':'.join(f'{c:02x}' for c in hwaddr)

def validate_options(ctx: 'Context'):
    """validate and filter options based on requested operation-codes"""
    options   = OptionList()
    requested = set(ctx.request.requested_options())
    for option in ctx.response.options:
        if option.opcode in requested or option.opcode in REQUIRED:
            options.append(option)
    options.sort()
    ctx.response.options = options

#** Classes **#

@dataclass
class Session(BaseSession, ABC):
    logger: Logger = logger

    ## Overrides

    @abstractmethod
    def process_discover(self, req: Message) -> Optional[Message]:
        raise NotImplementedError 

    @abstractmethod
    def process_request(self, req: Message) -> Optional[Message]:
        raise NotImplementedError 

    @abstractmethod
    def process_decline(self, req: Message) -> Optional[Message]:
        raise NotImplementedError 

    @abstractmethod
    def process_release(self, req: Message) -> Optional[Message]:
        raise NotImplementedError 

    @abstractmethod
    def process_inform(self, req: Message) -> Optional[Message]:
        raise NotImplementedError

    @abstractmethod
    def process_unknown(self, req: Message) -> Optional[Message]:
        raise NotImplementedError

    def default_response(self, req: Message) -> Message:
        """Generate Default DHCP Response"""
        return Message(
            op=OpCode.BootReply,
            id=req.id,
            client_hw=req.client_hw,
            options=OptionList(),
        )
 
    def send(self, request: Message, response: Message):
        """send completed response message"""
        hosts = (
            request.gateway_addr,
            request.client_addr,
            IPv4Address(self.addr.host),
            BROADCAST,
        )
        # encode and send response
        data = response.encode()
        for ipv4 in hosts:
            if is_zero(ipv4):
                continue
            host = str(ipv4)
            info = f'{self.addr_str} | sent {len(data)} bytes to {host}:{PORT}'
            self.logger.debug(info)
            self.writer.write(data, addr=(host, PORT))

    ## Session Impl

    def connection_made(self, addr: Address, writer: UdpWriter):
        """handle session initialization on connection-made"""
        self.addr      = addr
        self.writer    = writer
        self.addr_str  = '%s:%d' % self.addr

    def data_recieved(self, data: bytes):
        """recieve DHCPv4 request and generate response"""
        self.logger.debug(f'{self.addr_str} | recieved {len(data)} bytes')
        # parse initial request message
        request      = Message.decode(data)
        message_type = request.message_type()
        if message_type is None:
            self.logger.warning(f'{self.addr_str} | no message-type specified')
        # send to relevant handler
        response: Optional[Message] = None
        try:
            if message_type == MessageType.Discover:
                response = self.process_discover(request)
            elif message_type == MessageType.Request:
                response = self.process_request(request)
            elif message_type == MessageType.Decline:
                response = self.process_decline(request)
            elif message_type == MessageType.Release:
                response = self.process_release(request)
            elif message_type == MessageType.Inform:
                response = self.process_inform(request)
            else:
                response = self.process_unknown(request)
        except DhcpError as e:
            nak      = OptMessageType(MessageType.Nak)
            status   = OptStatusCode(e.code, str(e).encode())
            response = response or self.default_response(request)
            response.options.extend([nak, status])
        except Exception as e:
            nak      = OptMessageType(MessageType.Nak)
            status   = OptStatusCode(StatusCode.UnspecFail, str(e).encode())
            response = response or self.default_response(request)
            response.options.extend([nak, status])
            raise e
        finally:
            if response is not None:
                self.send(request, response)
            else:
                self.logger.error(f'{self.addr_str} | No Response Given.')
                self.writer.close()

    def connection_lost(self, err: Optional[Exception]):
        """debug log connection lost"""
        msg = f'{self.addr_str} | connection-lost err={err}'
        log = self.logger.warning if err is not None else self.logger.debug
        log(msg)

class Context(NamedTuple):
    """Session Context to Pass to Handler Functions"""
    session:  Session
    request:  Message
    response: Message

@dataclass(slots=True)
class SimpleSession(Session):
    """Simplified DHCP Server Implementation"""
    logger:    Logger
    server_id: IPv4Address
    handlers:  List[HandlerFunc] = field(default_factory=list)
    releasers: List[HandlerFunc] = field(default_factory=list)
 
    def __post_init__(self):
        if not self.handlers:
            raise ValueError('SimpleSession Handlers Empty!')

    def process_discover(self, req: Message) -> Optional[Message]:
        """process incoming discover request"""
        msg = OptMessageType(MessageType.Offer)
        res = self.default_response(req)
        ctx = Context(self, req, res)
        res.server_addr = self.server_id
        for func in self.handlers:
            func(ctx)
        res.options.setdefault(OptServerId(self.server_id))
        res.options.setdefault(msg)
        validate_options(ctx)
        res.options.move_to_start(msg.opcode)
        # ensure your-ip is set during exchange
        if not res.your_addr:
            raise RuntimeError(f'Handlers Failed to Assign Client-IP')
        return res

    def process_request(self, req: Message) -> Optional[Message]:
        """process incoming dhcp request"""
        msg = OptMessageType(MessageType.Ack)
        res = self.default_response(req)
        ctx = Context(self, req, res)
        res.server_addr = self.server_id
        for handler in self.handlers:
            handler(ctx)
        res.options.setdefault(OptServerId(self.server_id))
        res.options.setdefault(msg)
        validate_options(ctx)
        res.options.move_to_start(msg.opcode)
        # ensure your-ip is set during exchange
        if not res.your_addr:
            raise RuntimeError(f'Handlers Failed to Assign Client-IP')
        # ensure assignment matches request
        netmask  = res.options.get(OptSubnetMask.opcode)
        netmask  = netmask.value if netmask else None
        req_addr = req.requested_address()
        req_cast = req.broadcast_address()
        conflict = req_addr and req_addr != res.your_addr
        conflict = conflict or req_cast and req_cast != netmask
        if conflict:
            self.logger.warning(f'{self.addr_str} | REQUEST NAK {req_addr!r}')
            nak = OptMessageType(MessageType.Nak)
            res.options.set(nak)
        return res

    def process_decline(self, req: Message) -> Optional[Message]:
        """process incoming decline response"""
        mac = mac_address(req.client_hw)
        res = self.default_response(req)
        msg = OptMessageType(MessageType.Nak)
        ctx = Context(self, req, res)
        res.server_addr = self.server_id
        self.logger.info(f'{self.addr_str} | DECLINE {mac} {req.client_addr}')
        for func in self.releasers:
            func(ctx)
        res.options.setdefault(OptServerId(self.server_id))
        res.options.setdefault(msg)
        validate_options(ctx)
        res.options.move_to_start(msg.opcode)
        return res

    def process_release(self, req: Message) -> Optional[Message]:
        """process incoming release request"""
        mac = mac_address(req.client_hw)
        res = self.default_response(req)
        msg = OptMessageType(MessageType.Ack)
        ctx = Context(self, req, res)
        res.server_addr = self.server_id
        self.logger.info(f'{self.addr_str} | RELEASE {mac} {req.client_addr}')
        for func in self.releasers:
            func(ctx)
        res.options.setdefault(OptServerId(self.server_id))
        res.options.setdefault(msg)
        validate_options(ctx)
        res.options.move_to_start(msg.opcode)
        return res

    def process_inform(self, _: Message) -> Optional[Message]:
        """process incoming inform request"""
        raise NotAllowed('Inform not Allowed')
 
    def process_unknown(self, _: Message) -> Optional[Message]:
        raise UnknownQueryType('Unsupported MessageType')



"""
BaseClass ABC Protocol Implementations for DHCP v4/v6
"""
from enum import IntEnum
from datetime import timedelta
from typing import *

from pystructs import Struct, GreedyBytes

#** Variables **#
__all__ = ['Seconds', 'Microseconds', 'DHCPOption', 'DHCPOptionList']

#** Classes **#

class Timedelta(timedelta):
    """Integer Codec Compatable Timedelta"""
    field: str = 'seconds'

    def __class_getitem__(cls, field: str):
        return type(field.title(), (cls, ), {'field': field})

    def __new__(cls, delta: SupportsInt = 0):
        return super().__new__(cls, **{cls.field: int(delta)})

    def __int__(self) -> int:
        return int(self.total_seconds())

Seconds      = Timedelta['seconds']
Microseconds = Timedelta['microseconds']

class DHCPOption(Struct):
    opcode: ClassVar[IntEnum]
    value:  GreedyBytes

class DHCPOptionList:
    """Hybrid Between Dict/List for Quick Option Collection"""

    def __init__(self, data: Optional[List[DHCPOption]] = None):
        self.data    = []
        self.opcodes = set()
        self.extend(data or [])
 
    def keys(self) -> Iterator[IntEnum]:
        return (option.opcode for option in self.data)

    def values(self) -> Iterator[DHCPOption]:
        return (option for option in self.data)

    def items(self) -> Iterator[Tuple[IntEnum, DHCPOption]]:
        return iter((option.opcode, option) for option in self.data)

    def __repr__(self) -> str:
        return repr(self.data)

    def __iter__(self) -> Iterator[DHCPOption]:
        yield from self.values()

    def __contains__(self, key: Union[IntEnum, DHCPOption]) -> bool:
        key = key.opcode if isinstance(key, DHCPOption) else key
        return key in self.opcodes

    def __getitem__(self, op: IntEnum):
        if op not in self.opcodes:
            raise KeyError(op)
        for option in self.data:
            if option.opcode == op:
                return option
        raise KeyError(op)

    def __setitem__(self, op: IntEnum, value: DHCPOption):
        if op not in self.opcodes:
            self.opcodes.add(op)
            self.data.append((op, value))
            return
        for n, option in enumerate(self.data, 0):
            if option.opcode == op:
                self.data[n] = value
                return
        raise AttributeError(f'Failed to set {op!r}')

    def get(self, key: IntEnum, default: Any = None):
        if key not in self.opcodes:
            return default
        return self[key]

    def set(self, option: DHCPOption):
        self[option.opcode] = option
 
    def setdefault(self, option: DHCPOption):
        if option.opcode not in self.opcodes:
            self.append(option)

    def sort(self):
        self.data.sort(key=lambda option: option.opcode)

    def append(self, option: DHCPOption):
        if option.opcode in self.opcodes:
            raise ValueError(f'Option: {option.opcode!r} already present')
        self.data.append(option)
        self.opcodes.add(option.opcode)
 
    def extend(self, options: Sequence[DHCPOption]):
        for opt in options:
            self.append(opt)

    def insert(self, pos: int, option: DHCPOption):
        if option.opcode in self.opcodes:
            raise ValueError(f'Option: {option.opcode} already present')
        self.data.insert(pos, option)
        self.opcodes.add(option.opcode)
 
    def index(self, key: IntEnum):
        if key not in self.opcodes:
            raise ValueError(f'Option: {key} not present')
        for n, opcode in enumerate(self.keys(), 0):
            if opcode == key:
                return n
        raise ValueError(f'Failed to Index: {key}')

    def move_to_start(self, key: IntEnum):
        index = self.index(key)
        item  = self.data.pop(index)
        self.opcodes.remove(key)
        self.insert(0, item)



"""
DHCPv4 Option Implementations
"""
from typing import ClassVar, List, SupportsInt
from typing_extensions import Annotated
from warnings import warn

from pystructs import *
from pyderive import dataclass

from .enum import OptionCode, MessageType
from ..enum import Arch, StatusCode
from ..base import DHCPOption, DHCPOptionList, Seconds

#** Variables **#
__all__ = [
    'read_option',

    'Option',
    'OptionList',
    'OptEnd',
    'OptServerId',
    'OptSubnetMask',
    'OptBroadcast',
    'OptRouter',
    'OptDNS',
    'OptRequestedAddr',
    'OptStatusCode',
    'OptIPLeaseTime',
    'OptRenwalTime',
    'OptRebindTime',
    'OptMessageType',
    'OptParamRequestList',
    'OptMaxMessageSize',
    'OptClassIdentifier',
    'OptTFTPServerName',
    'OptTFTPServerIP',
    'OptBootFile',
    'OptPXEPathPrefix',
    'OptUserClassInfo',
    'OptClientSystemArch',
    'OptClientNetworkIface',
    'OptClientMachineID',
    'OptEtherBoot',
]

#: codec to parse Int8 as OptionCode
OptionCodeInt = Annotated[OptionCode, Wrap[U8, OptionCode]]

#** Functions **#

def read_option(ctx: Context, raw: bytes) -> 'Option':
    """
    read and deserialize best option-class to match option content
    
    :param ctx: serialization context
    :param raw: raw bytes to deserialized into an option
    :return:    option object
    """
    # parse option-type/value & retrieve best matching subclass
    opt     = OptStruct.decode(ctx, raw)
    oclass  = OPTIONS.get(opt.code, Option)
    # parse option w/ it's own sub-context
    subctx = Context()
    option = oclass.decode(subctx, opt.value)
    if oclass is Option:
        oclass.opcode = opt.code
    return option

def write_option(ctx: Context, option) -> bytes:
    """
    write and serialize option-class into raw-bytes
    
    :param ctx: serialization context
    :param raw: option object to serialize
    :return:    serialized bytes
    """
    # serialize option into bytes
    subctx  = Context()
    content = option.encode(subctx)
    return OptStruct(option.opcode, content).encode(ctx)

#** Classes **#

class OptStruct(Struct):
    code:  OptionCodeInt
    value: Annotated[bytes, SizedBytes[U8]]
 
    def __post_init__(self):
        """warn when value is too long and truncate"""
        if len(self.value) <= 255:
            return
        code       = self.code
        vlen       = len(self.value)
        self.value = self.value[:252] + b'...'
        warn(f'{code!r} value too long {vlen}. truncating!', RuntimeWarning)

@dataclass
class Option(DHCPOption):
    """DHCPv4 BaseClass Option Definition"""
    opcode: ClassVar[OptionCode] = OptionCode.OptionPad

class OptionList(DHCPOptionList):
    """DHCPv4 DHCPOptionList Implementation"""
    pass

### SubClasses

class OptEnd(Option):
    opcode = OptionCode.End

class OptHostName(Option):
    opcode = OptionCode.HostName

class _Ipv4Option(Option):
    opcode: ClassVar[OptionCode]
    value:  IPv4

class OptServerId(_Ipv4Option):
    opcode = OptionCode.ServerIdentifier

class OptSubnetMask(_Ipv4Option):
    opcode = OptionCode.SubnetMask

class OptBroadcast(_Ipv4Option):
    opcode = OptionCode.BroadcastAddress

class OptRouter(_Ipv4Option):
    opcode = OptionCode.Router

class OptDNS(_Ipv4Option):
    opcode = OptionCode.DomainNameServer
    value: Annotated[List[Ipv4Type], GreedyList[IPv4]]

class OptRequestedAddr(_Ipv4Option):
    opcode = OptionCode.RequestedIPAddress

class OptStatusCode(Option):
    opcode = OptionCode.StatusCode
    value:   Annotated[StatusCode, Wrap[U8, StatusCode]]
    message: GreedyBytes

class OptIPLeaseTime(Option):
    opcode = OptionCode.IPAddressLeaseTime
    value: Annotated[SupportsInt, Wrap[U32, Seconds]]

class OptRenwalTime(Option):
    opcode = OptionCode.RenewTimeValue
    value: Annotated[SupportsInt, Wrap[U32, Seconds]]

class OptRebindTime(Option):
    opcode = OptionCode.RenewTimeValue
    value: Annotated[SupportsInt, Wrap[U32, Seconds]]

class OptMessageType(Option):
    opcode = OptionCode.DHCPMessageType
    value: Annotated[MessageType, Wrap[U8, MessageType]]

class OptParamRequestList(Option):
    opcode = OptionCode.ParameterRequestList
    value: Annotated[OptionCode, GreedyList[OptionCodeInt]]

class OptMaxMessageSize(Option):
    opcode = OptionCode.MaximumDHCPMessageSize
    value: U16

class OptClassIdentifier(Option):
    opcode = OptionCode.ClassIdentifier

class OptTFTPServerName(Option):
    opcode = OptionCode.TFTPServerName

class OptTFTPServerIP(_Ipv4Option):
    opcode = OptionCode.TFTPServerIPAddress

class OptBootFile(Option):
    opcode = OptionCode.BootfileName

class OptPXEPathPrefix(Option):
    opcode = OptionCode.PXELinuxPathPrefix

class OptUserClassInfo(Option):
    opcode = OptionCode.UserClassInformation

class OptClientSystemArch(Option):
    opcode = OptionCode.ClientSystemArchitectureType
    value: Annotated[List[Arch], GreedyList[Wrap[U16, Arch]]]

class OptClientNetworkIface(Option):
    opcode = OptionCode.ClientNetworkInterfaceIdentifier
    value: Const[b'\x01']
    major: U8
    minor: U8

class OptClientMachineID(Option):
    opcode = OptionCode.ClientMachineIdentifier
    value:  GreedyBytes

class OptEtherBoot(Option):
    opcode = OptionCode.Etherboot

#** Init **#

#: map of option-codes to option-subclasses
OPTIONS = {
    value.opcode:value
    for value in globals().values()
    if isinstance(value, type) and issubclass(value, Option) 
}



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    readme = f.read()

setup(
    name='pydhcp3',
    version='0.0.5',
    license='MIT',
    author='Andrew Scott',
    author_email='imgurbot12@gmail.com',
    url='https://github.com/imgurbot12/pydhcp',
    description="Simple Python DHCP Library. DHCP Packet-Parsing/Client/Server",
    long_description=readme,
    long_description_content_type="text/markdown",
    python_requires='>=3.6',
    packages=find_packages(),
    install_requires=[
        'pyserve3',
        'pyderive3',
        'pystructs3',
        'typing_extensions'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



"""
SimpleSession Handler Implentations
"""
from copy import copy
from enum import IntEnum
from functools import wraps
from datetime import datetime, timedelta
from ipaddress import IPv4Address, IPv4Interface
from typing import *
from threading import Lock

from pyderive import dataclass, field

from .server import HandlerFunc, Context
from ..message import Message, OptionCode
from ..options import *
from ...enum import Arch

#** Variables **#
__all__ = [
    'pxe_handler', 
    'ipv4_handler',

    'AssignFunc',
    'Assignment',
    'PxeTftpConfig',
    'PxeDynConfig',
    'PxeConfig',
    'Cache',
]

#: typehint for optional cache
OptCache = Optional['Cache']

#: typehint definition for assignment function
AssignFunc = Callable[[str], 'Assignment']

#: set of pxe supported option-codes
PXE_OPTIONS = {
    OptionCode.TFTPServerName,
    OptionCode.TFTPServerAddress,
    OptionCode.TFTPServerIPAddress,
    OptionCode.BootfileName,
    OptionCode.PXELinuxPathPrefix,
}

#** Functions **#

def _getopt(req: Message, opcode: OptionCode) -> Any:
    """retrieve bytes associated w/ opcode if in request else none"""
    op = req.options.get(opcode)
    return op.value if op else None

def _enum(value: Union[int, str, IntEnum], enum: Type[IntEnum]):
    """convert value to int-enum"""
    if isinstance(value, int):
        return enum(value)
    elif isinstance(value, str):
        return enum[value]
    elif isinstance(value, enum):
        return value
    else:
        raise ValueError(f'Invalid {enum}: {value!r}')

def pxe_handler(rootcfg: 'PxeConfig') -> HandlerFunc:
    """
    generate session handler for dhcp pxe assignment

    :param rootcfg: root pxe configuration settings
    :return:        session handler function
    """
    def assign_pxe(ctx: Context):
        """
        DHCP PXE Assignment Function

        :param ctx: request/response context
        """
        sess, req, res = ctx
        # skip any assignment if no options are requested
        options = set(req.requested_options())
        if not any(op for op in PXE_OPTIONS if op in options):
            return
        # determine config based on default or or arch/vendor specific
        hwaddr = ctx.request.client_hw.hex()
        config = rootcfg
        subcfg = None
        arches = _getopt(req, OptionCode.ClientSystemArchitectureType)
        vendor = _getopt(req, OptionCode.ClassIdentifier)
        if not subcfg and arches and config.dynamic:
            sess.logger.debug(f'{hwaddr} arches={arches!r}')
            for arch in arches:
                subcfg = config.dynamic.arches.get(arch)
                if subcfg:
                    break
        if not subcfg and vendor and config.dynamic:
            vendor = vendor.decode()
            sess.logger.debug(f'{hwaddr} vendor={vendor!r}')
            for vid, match in config.dynamic.vendors.items():
                if match in vendor:
                    subcfg = config.dynamic.configs.get(vid)
                    sess.logger.debug(f'{hwaddr} vendor match {vid} {match}')
                    if subcfg:
                        break
        # build config from subconfig
        if subcfg:
            config = copy(config)
            config.ipaddr   = subcfg.ipaddr or config.ipaddr
            config.hostname = subcfg.hostname or config.hostname
            config.filename = subcfg.filename or config.filename
        # build options based on config and logging message
        message = [f'{sess.addr_str} | {req.client_hw.hex()}']
        message.append(f'-> pxe={config.ipaddr}')
        res.server_addr = config.ipaddr
        res.options.append(OptTFTPServerIP(config.ipaddr))
        if config.primary:
            res.boot_file   = config.filename or res.boot_file
            res.server_name = config.hostname or res.server_name
        if config.prefix:
            res.options.append(OptPXEPathPrefix(config.prefix))
            message.append(f'root={config.prefix.decode()!r}')
        if config.hostname:
            res.options.append(OptTFTPServerName(config.hostname))
            message.append(f'host={config.hostname.decode()!r}')
        if config.filename:
            res.options.append(OptBootFile(config.filename))
            message.append(f'file={config.filename.decode()!r}')
        sess.logger.info(' '.join(message))
    return assign_pxe

def ipv4_handler(func: AssignFunc, cache: OptCache = None) -> HandlerFunc:
    """
    generate session handler for dhcp ipv4 assignment

    :param func:  assignment function
    :param cache: optional cache to use when generating assignments
    :return:      session handler function
    """
    @wraps(func)
    def assign_ipv4(ctx: Context):
        """
        DHCP IP-LEASE Assignment Function

        :param ctx: request/response context
        """
        sess, req, res = ctx
        hwaddr = req.client_hw.hex()
        source = 'CACHE'
        assign = cache.get(hwaddr) if cache else None
        if not assign:
            source = 'LOOKUP'
            assign = func(hwaddr)
            if cache:
                cache.put(hwaddr, assign)
        # log assignment
        dns = ','.join(str(dns) for dns in assign.dns)
        sess.logger.info(
            f'{sess.addr_str} | {hwaddr} -> ip={assign.ipaddr} ' + \
            f'gw={assign.gateway} dns={dns} lease={assign.lease} src={source}')
        # assign data to response
        res.your_addr = assign.ipaddr.ip
        res.options.extend([
            OptDNS(assign.dns),
            OptRouter(assign.gateway),
            OptSubnetMask(assign.ipaddr.netmask),
            OptIPLeaseTime(int(assign.lease.total_seconds())),
        ])
    return assign_ipv4

#** Classes **#

class Assignment(NamedTuple):
    """IP Assignment Function"""
    dns:     List[IPv4Address]
    ipaddr:  IPv4Interface
    gateway: IPv4Address
    lease:   timedelta

@dataclass
class PxeTftpConfig:
    """Dynamic TFTP PXE Configuration"""
    filename: bytes
    hostname: Optional[bytes]       = None
    ipaddr:   Optional[IPv4Address] = None

ArchT = Union[int, str, Arch]

@dataclass
class PxeDynConfig:
    """Dynamic PXE Configuration Settings and Translations"""
    vendors: Dict[str, str]             = field(default_factory=dict)
    arches:  Dict[ArchT, PxeTftpConfig] = field(default_factory=dict)
    configs: Dict[str, PxeTftpConfig]   = field(default_factory=dict)

    def __post_init__(self):
        """validate and ensure arches are arch-enums"""
        self.arches  = {_enum(k, Arch):v for k, v in self.arches.items()}

@dataclass
class PxeConfig:
    """DHCP PXE Confguration Settings"""
    ipaddr:   IPv4Address
    primary:  bool            = False
    prefix:   Optional[bytes] = None
    hostname: Optional[bytes] = None
    filename: Optional[bytes] = None
    dynamic:  PxeDynConfig    = field(default_factory=PxeDynConfig)

class CacheRecord(NamedTuple):
    """Cache Record Instance"""
    assign:     Any
    expiration: Optional[datetime]

class Cache:
    """ThreadSafe Expiring Cache Implementation"""
    __slots__ = ('expiration', 'maxsize', 'mutex', 'cache')

    def __init__(self, 
        expiration: Optional[timedelta] = None, 
        maxsize:    Optional[int]       = None,
    ):
        self.expiration = expiration
        self.maxsize    = maxsize
        self.mutex      = Lock()
        self.cache: Dict[str, CacheRecord] = {}

    def get(self, key: str) -> Any:
        """retrieve non-expired key from cache"""
        if key not in self.cache:
            return
        with self.mutex:
            record = self.cache[key]
            if record.expiration and record.expiration <= datetime.now():
                del self.cache[key]
                return
            return record.assign

    def put(self, key: str, value: Any):
        """save new assignment to cache"""
        with self.mutex:
            # clear cache until under max-size
            while self.maxsize and len(self.cache) >= self.maxsize:
                self.cache.popitem()
            # assign expiration
            expiration = None
            if self.expiration:
                expiration = datetime.now() + self.expiration
            self.cache[key] = CacheRecord(value, expiration)

    def delete(self, key: str):
        """delete potential key from cache"""
        if key in self.cache:
            with self.mutex:
                self.cache.pop(key)
